from torchvision.datasets.utils import download_url, makedir_exist_ok

raw_folder = '.data/EMNIST/raw'

url = 'http://www.itl.nist.gov/iaui/vip/cs_links/EMNIST/gzip.zip'
md5 = "58c8d27c78d21e728a6bc7b3cc06412e"

makedir_exist_ok(raw_folder)

# download files
print('Downloading zip archive')
download_url(url, root=raw_folder, filename="emnist.zip", md5=md5)



import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms

import pytorch_warmup as warmup
import os


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4*4*50, 500)
        self.fc2 = nn.Linear(500, 47)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4*4*50)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


def train(args, model, device, train_loader, optimizer, lr_scheduler,
          warmup_scheduler, epoch, history):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        lr = optimizer.param_groups[0]['lr']
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        with warmup_scheduler.dampening():
            lr_scheduler.step()
        if (batch_idx+1) % args.log_interval == 0:
            loss = loss.item()
            step = warmup_scheduler.last_step
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f} LR: {:.6f}'.format(
                epoch, (batch_idx+1) * len(data), len(train_loader) * len(data),
                100. * (batch_idx+1) / len(train_loader), loss, lr))
            history.write(f'{epoch},{step},{loss},{lr}\n')


def test(args, model, device, test_loader, epoch, evaluation):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item()  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)
    test_acc = 100. * correct / len(test_loader.dataset)
    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
        test_loss, correct, len(test_loader.dataset), test_acc))
    evaluation.write(f'{epoch},{test_loss},{test_acc}\n')
    evaluation.flush()


def main():
    # Training settings
    parser = argparse.ArgumentParser(description='PyTorch EMNIST Example')
    parser.add_argument('--batch-size', type=int, default=64, metavar='N',
                        help='input batch size for training (default: 64)')
    parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--epochs', type=int, default=10, metavar='N',
                        help='number of epochs to train (default: 10)')
    parser.add_argument('--lr', type=float, default=0.01, metavar='LR',
                        help='base learning rate (default: 0.01)')
    parser.add_argument('--lr-min', type=float, default=1e-5, metavar='LM',
                        help='minimum learning rate (default: 1e-5)')
    parser.add_argument('--wd', type=float, default=0.01, metavar='WD',
                        help='weight decay (default: 0.01)')
    parser.add_argument('--beta2', type=float, default=0.999, metavar='B2',
                        help="Adam's beta2 parameter (default: 0.999)")
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')
    parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                        help='how many batches to wait before logging training status')
    parser.add_argument('--warmup', type=str, default='linear',
                        choices=['linear', 'exponential', 'radam', 'none'],
                        help='warmup schedule')
    parser.add_argument('--save-model', action='store_true', default=False,
                        help='For Saving the current Model')
    args = parser.parse_args()
    use_cuda = not args.no_cuda and torch.cuda.is_available()

    torch.manual_seed(args.seed)

    device = torch.device("cuda" if use_cuda else "cpu")

    kwargs = {'num_workers': 1, 'pin_memory': True} if use_cuda else {}
    train_loader = torch.utils.data.DataLoader(
        datasets.EMNIST('.data', 'balanced', train=True, download=True,
                        transform=transforms.Compose([
                            transforms.ToTensor(),
                            transforms.Normalize((0.1751,), (0.3332,))
                        ])),
        batch_size=args.batch_size, shuffle=True, drop_last=True, **kwargs)
    test_loader = torch.utils.data.DataLoader(
        datasets.EMNIST('.data', 'balanced', train=False,
                        transform=transforms.Compose([
                            transforms.ToTensor(),
                            transforms.Normalize((0.1751,), (0.3332,))
                        ])),
        batch_size=args.test_batch_size, shuffle=False, **kwargs)

    output_dir = args.warmup
    os.makedirs(output_dir, exist_ok=True)

    history = open(os.path.join(output_dir, 'history.csv'), 'w')
    history.write('epoch,step,loss,lr\n')

    evaluation = open(os.path.join(output_dir, 'evaluation.csv'), 'w')
    evaluation.write('epoch,loss,accuracy\n')

    model = Net().to(device)

    optimizer = optim.AdamW(model.parameters(), lr=args.lr,
                            betas=(0.9, args.beta2),
                            weight_decay=args.wd)
    num_steps = len(train_loader) * args.epochs
    lr_scheduler = optim.lr_scheduler.CosineAnnealingLR(
        optimizer, T_max=num_steps, eta_min=args.lr_min)
    if args.warmup == 'linear':
        warmup_scheduler = warmup.UntunedLinearWarmup(optimizer)
    elif args.warmup == 'exponential':
        warmup_scheduler = warmup.UntunedExponentialWarmup(optimizer)
    elif args.warmup == 'radam':
        warmup_scheduler = warmup.RAdamWarmup(optimizer)
    elif args.warmup == 'none':
        warmup_scheduler = warmup.LinearWarmup(optimizer, 1)

    for epoch in range(1, args.epochs + 1):
        train(args, model, device, train_loader, optimizer, lr_scheduler,
              warmup_scheduler, epoch, history)
        test(args, model, device, test_loader, epoch, evaluation)

    if (args.save_model):
        torch.save(model.state_dict(), os.path.join(output_dir, "emnist_cnn.pt"))

    history.close()
    evaluation.close()


if __name__ == '__main__':
    main()



import argparse
import matplotlib.pyplot as plt
import torch
from pytorch_warmup import *


def get_rates(warmup_cls, beta2, max_step):
    rates = []
    p = torch.nn.Parameter(torch.arange(10, dtype=torch.float32))
    optimizer = torch.optim.Adam([{'params': p}], lr=1.0, betas=(0.9, beta2))
    lr_scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda step: 1.0)
    warmup_scheduler = warmup_cls(optimizer)
    for step in range(1, max_step+1):
        rates.append(optimizer.param_groups[0]['lr'])
        optimizer.zero_grad()
        optimizer.step()
        lr_scheduler.step()
        warmup_scheduler.dampen()
    return rates


parser = argparse.ArgumentParser(description='Warmup schedule')
parser.add_argument('--output', type=str, default='none',
                    choices=['none', 'png', 'pdf'],
                    help='Output file type (default: none)')
args = parser.parse_args()

beta2 = 0.999
max_step = 3000

plt.plot(range(1, max_step+1), get_rates(RAdamWarmup, beta2, max_step), label='RAdam')
plt.plot(range(1, max_step+1), get_rates(UntunedExponentialWarmup, beta2, max_step), label='Untuned Exponential')
plt.plot(range(1, max_step+1), get_rates(UntunedLinearWarmup, beta2, max_step), label='Untuned Linear')
plt.legend()
plt.title('Warmup Schedule')
plt.xlabel('Iteration')
plt.ylabel(r'Warmup factor $(\omega_t)$')
if args.output == 'none':
    plt.show()
else:
    plt.savefig(f'warmup_schedule.{args.output}')



import math
from .base import BaseWarmup


def rho_inf_fn(beta2):
    return 2.0 / (1 - beta2) - 1


def rho_fn(t, beta2, rho_inf):
    b2t = beta2 ** t
    rho_t = rho_inf - 2 * t * b2t / (1 - b2t)
    return rho_t


def get_offset(beta2, rho_inf):
    if not beta2 > 0.6:
        raise ValueError('beta2 ({}) must be greater than 0.6'.format(beta2))
    offset = 1
    while True:
        if rho_fn(offset, beta2, rho_inf) > 4:
            return offset
        offset += 1


class RAdamWarmup(BaseWarmup):
    """RAdam warmup schedule.

    This warmup scheme is described in
    `On the adequacy of untuned warmup for adaptive optimization
    <https://arxiv.org/abs/1910.04209>`_.

    Arguments:
        optimizer (Optimizer): an Adam optimizer
        last_step (int): The index of last step. (Default: -1)
    """

    def __init__(self, optimizer, last_step=-1):
        warmup_params = [
            dict(
                beta2=x['betas'][1],
                rho_inf=rho_inf_fn(x['betas'][1]),
            )
            for x in optimizer.param_groups
        ]
        for x in warmup_params:
            x['offset'] = get_offset(**x)
        super(RAdamWarmup, self).__init__(optimizer, warmup_params, last_step)

    def warmup_factor(self, step, beta2, rho_inf, offset):
        rho = rho_fn(step+offset, beta2, rho_inf)
        numerator = (rho - 4) * (rho - 2) * rho_inf
        denominator = (rho_inf - 4) * (rho_inf - 2) * rho
        return math.sqrt(numerator/denominator)



import argparse
import matplotlib.pyplot as plt
from pytorch_warmup import *
import numpy as np


def untuned_exponential_period(beta2):
    return 1.0 / (np.exp(1.0 - beta2) - 1.0)


def untuned_linear_period(beta2):
    return 1.0 / (1.0 - beta2) - 0.5


def warmup_factor(step, beta2, rho_inf, offset):
    rho = rho_fn(step+offset, beta2, rho_inf)
    numerator = (rho - 4) * (rho - 2) * rho_inf
    denominator = (rho_inf - 4) * (rho_inf - 2) * rho
    return np.sqrt(numerator/denominator)


def radam_period_fn(beta2):
    rho_inf = rho_inf_fn(beta2)
    offset = get_offset(beta2, rho_inf)
    steps = np.arange(1, 101)
    w = warmup_factor(steps, beta2, rho_inf, offset)
    total_sum = np.sum(1-w)
    t = 1
    while True:
        steps = np.arange(100*t+1, 100*(t+1)+1)
        w = warmup_factor(steps, beta2, rho_inf, offset)
        partial_sum = np.sum(1-w)
        if partial_sum < 0.1:
            break
        total_sum += partial_sum
        t += 1
    return total_sum


def radam_period(beta2):
    return [radam_period_fn(x) for x in beta2]


parser = argparse.ArgumentParser(description='Effective warmup period')
parser.add_argument('--output', type=str, default='none',
                    choices=['none', 'png', 'pdf'],
                    help='Output file type (default: none)')
args = parser.parse_args()

beta2 = np.arange(0.99, 0.9999, 0.0001)
plt.plot(beta2, untuned_exponential_period(beta2), label='Untuned Exponential')
plt.plot(beta2, untuned_linear_period(beta2), linestyle=':', label='Untuned Linear')
plt.plot(beta2, radam_period(beta2), linestyle='--', label='RAdam')
plt.xlim(0.990, 1.00)
plt.ylim(100, 10000)
plt.yscale('log')
plt.legend()
plt.title('Effective Warmup Period')
plt.xlabel(r'$\beta_{2}$')
plt.ylabel(r'${\cal T}(\omega)$')
if args.output == 'none':
    plt.show()
else:
    plt.savefig(f'warmup_period.{args.output}')



import math
from contextlib import contextmanager
from torch.optim import Optimizer


class BaseWarmup(object):
    """Base class for all warmup schedules

    Arguments:
        optimizer (Optimizer): an instance of a subclass of Optimizer
        warmup_params (list): warmup paramters
        last_step (int): The index of last step. (Default: -1)
    """

    def __init__(self, optimizer, warmup_params, last_step=-1):
        if not isinstance(optimizer, Optimizer):
            raise TypeError('{} is not an Optimizer'.format(
                type(optimizer).__name__))
        self.optimizer = optimizer
        self.warmup_params = warmup_params
        self.last_step = last_step
        self.lrs = [group['lr'] for group in self.optimizer.param_groups]
        self.dampen()

    def state_dict(self):
        """Returns the state of the warmup scheduler as a :class:`dict`.

        It contains an entry for every variable in self.__dict__ which
        is not the optimizer.
        """
        return {key: value for key, value in self.__dict__.items() if key != 'optimizer'}

    def load_state_dict(self, state_dict):
        """Loads the warmup scheduler's state.

        Arguments:
            state_dict (dict): warmup scheduler state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        self.__dict__.update(state_dict)

    def dampen(self, step=None):
        """Dampen the learning rates.

        Arguments:
            step (int): The index of current step. (Default: None)
        """
        if step is None:
            step = self.last_step + 1
        self.last_step = step

        for group, params in zip(self.optimizer.param_groups, self.warmup_params):
            omega = self.warmup_factor(step, **params)
            group['lr'] *= omega

    @contextmanager
    def dampening(self):
        for group, lr in zip(self.optimizer.param_groups, self.lrs):
            group['lr'] = lr
        yield
        self.lrs = [group['lr'] for group in self.optimizer.param_groups]
        self.dampen()

    def warmup_factor(self, step, **params):
        raise NotImplementedError


def get_warmup_params(warmup_period, group_count):
    if type(warmup_period) == list:
        if len(warmup_period) != group_count:
            raise ValueError(
                'size of warmup_period does not equal {}.'.format(group_count))
        for x in warmup_period:
            if type(x) != int:
                raise ValueError(
                    'An element in warmup_period, {}, is not an int.'.format(
                        type(x).__name__))
        warmup_params = [dict(warmup_period=x) for x in warmup_period]
    elif type(warmup_period) == int:
        warmup_params = [dict(warmup_period=warmup_period)
                         for _ in range(group_count)]
    else:
        raise TypeError('{} is not a list nor an int.'.format(
            type(warmup_period).__name__))
    return warmup_params


class LinearWarmup(BaseWarmup):
    """Linear warmup schedule.

    Arguments:
        optimizer (Optimizer): an instance of a subclass of Optimizer
        warmup_period (int or list): Warmup period
        last_step (int): The index of last step. (Default: -1)
    """

    def __init__(self, optimizer, warmup_period, last_step=-1):
        group_count = len(optimizer.param_groups)
        warmup_params = get_warmup_params(warmup_period, group_count)
        super(LinearWarmup, self).__init__(optimizer, warmup_params, last_step)

    def warmup_factor(self, step, warmup_period):
        return min(1.0, (step+1) / warmup_period)


class ExponentialWarmup(BaseWarmup):
    """Exponential warmup schedule.

    Arguments:
        optimizer (Optimizer): an instance of a subclass of Optimizer
        warmup_period (int or list): Effective warmup period
        last_step (int): The index of last step. (Default: -1)
    """

    def __init__(self, optimizer, warmup_period, last_step=-1):
        group_count = len(optimizer.param_groups)
        warmup_params = get_warmup_params(warmup_period, group_count)
        super(ExponentialWarmup, self).__init__(optimizer, warmup_params, last_step)

    def warmup_factor(self, step, warmup_period):
        return 1.0 - math.exp(-(step+1) / warmup_period)



from .base import LinearWarmup, ExponentialWarmup


class UntunedLinearWarmup(LinearWarmup):
    """Untuned linear warmup schedule for Adam.

    This warmup scheme is described in
    `On the adequacy of untuned warmup for adaptive optimization
    <https://arxiv.org/abs/1910.04209>`_.

    Arguments:
        optimizer (Optimizer): an Adam optimizer
        last_step (int): The index of last step. (Default: -1)
    """

    def __init__(self, optimizer, last_step=-1):
        def warmup_period_fn(beta2):
            return int(2.0 / (1.0-beta2))
        warmup_period = [warmup_period_fn(x['betas'][1]) for x in optimizer.param_groups]
        super(UntunedLinearWarmup, self).__init__(optimizer, warmup_period, last_step)


class UntunedExponentialWarmup(ExponentialWarmup):
    """Untuned exponetial warmup schedule for Adam.

    This warmup scheme is described in
    `On the adequacy of untuned warmup for adaptive optimization
    <https://arxiv.org/abs/1910.04209>`_.

    Arguments:
        optimizer (Optimizer): an Adam optimizer
        last_step (int): The index of last step. (Default: -1)
    """

    def __init__(self, optimizer, last_step=-1):
        def warmup_period_fn(beta2):
            return int(1.0 / (1.0-beta2))
        warmup_period = [warmup_period_fn(x['betas'][1]) for x in optimizer.param_groups]
        super(UntunedExponentialWarmup, self).__init__(optimizer, warmup_period, last_step)



import unittest
import math
import torch
import pytorch_warmup as warmup


def _test_state_dict(self, warmup_scheduler, constructor):
    warmup_scheduler_copy = constructor()
    warmup_scheduler_copy.load_state_dict(warmup_scheduler.state_dict())
    for key in warmup_scheduler.__dict__.keys():
        if key not in ['optimizer']:
            print(key, warmup_scheduler.__dict__[key])
            self.assertAlmostEqual(warmup_scheduler.__dict__[key],
                                   warmup_scheduler_copy.__dict__[key])


class TestBase(unittest.TestCase):

    def setUp(self):
        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    def test_linear(self):
        p1 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        p2 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        optimizer = torch.optim.SGD([
                {'params': [p1]},
                {'params': [p2], 'lr': 0.1}
            ], lr=0.5)
        lr_scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda step: 1.0)
        warmup_scheduler = warmup.LinearWarmup(optimizer, warmup_period=5)
        for step in range(1, 11):
            lr = [x['lr'] for x in optimizer.param_groups]
            print(f'{step} {lr}')
            if step < 5:
                self.assertAlmostEqual(lr[0], 0.5 * step / 5)
                self.assertAlmostEqual(lr[1], 0.1 * step / 5)
            else:
                self.assertAlmostEqual(lr[0], 0.5)
                self.assertAlmostEqual(lr[1], 0.1)
            optimizer.zero_grad()
            optimizer.step()
            with warmup_scheduler.dampening():
                lr_scheduler.step()

        _test_state_dict(self, warmup_scheduler,
                         lambda: warmup.LinearWarmup(optimizer, warmup_period=10))

    def test_exponetial(self):
        p1 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        p2 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        optimizer = torch.optim.SGD([
                {'params': [p1]},
                {'params': [p2], 'lr': 0.1}
            ], lr=0.5)
        lr_scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda step: 1.0)
        warmup_scheduler = warmup.ExponentialWarmup(optimizer, warmup_period=5)
        for step in range(1, 11):
            lr = [x['lr'] for x in optimizer.param_groups]
            print(f'{step} {lr}')
            self.assertAlmostEqual(lr[0], 0.5 * (1 - math.exp(-step / 5)))
            self.assertAlmostEqual(lr[1], 0.1 * (1 - math.exp(-step / 5)))
            optimizer.zero_grad()
            optimizer.step()
            with warmup_scheduler.dampening():
                lr_scheduler.step()

        _test_state_dict(self, warmup_scheduler,
                         lambda: warmup.ExponentialWarmup(optimizer, warmup_period=10))

    def test_linear_chaining(self):
        def preparation():
            p1 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
            p2 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
            optimizer = torch.optim.SGD([
                    {'params': [p1]},
                    {'params': [p2], 'lr': 0.1}
                    ], lr=0.5)
            lr_scheduler1 = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.9)
            lr_scheduler2 = torch.optim.lr_scheduler.StepLR(optimizer, step_size=3, gamma=0.1)
            return optimizer, lr_scheduler1, lr_scheduler2

        # First record undampened lrs
        optimizer, lr_scheduler1, lr_scheduler2 = preparation()
        lrs = []
        print()
        print('Undampened:')
        for step in range(1, 11):
            lr = [x['lr'] for x in optimizer.param_groups]
            print(f'{step} {lr}')
            lrs.append(lr)
            optimizer.zero_grad()
            optimizer.step()
            lr_scheduler1.step()
            lr_scheduler2.step()

        optimizer, lr_scheduler1, lr_scheduler2 = preparation()
        warmup_scheduler = warmup.LinearWarmup(optimizer, warmup_period=5)
        print('Dampened:')
        for step in range(1, 11):
            lr = [x['lr'] for x in optimizer.param_groups]
            print(f'{step} {lr}')
            if step < 5:
                self.assertAlmostEqual(lr[0], lrs[step-1][0] * step / 5)
                self.assertAlmostEqual(lr[1], lrs[step-1][1] * step / 5)
            else:
                self.assertAlmostEqual(lr[0], lrs[step-1][0])
                self.assertAlmostEqual(lr[1], lrs[step-1][1])
            optimizer.zero_grad()
            optimizer.step()
            with warmup_scheduler.dampening():
                lr_scheduler1.step()
                lr_scheduler2.step()

        _test_state_dict(self, warmup_scheduler,
                         lambda: warmup.LinearWarmup(optimizer, warmup_period=10))

    def test_linear_wo_lr_scheduler(self):
        p1 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        p2 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        optimizer = torch.optim.SGD([
                {'params': [p1]},
                {'params': [p2], 'lr': 0.1}
            ], lr=0.5)
        warmup_scheduler = warmup.LinearWarmup(optimizer, warmup_period=5)
        for step in range(1, 11):
            lr = [x['lr'] for x in optimizer.param_groups]
            print(f'{step} {lr}')
            if step < 5:
                self.assertAlmostEqual(lr[0], 0.5 * step / 5)
                self.assertAlmostEqual(lr[1], 0.1 * step / 5)
            else:
                self.assertAlmostEqual(lr[0], 0.5)
                self.assertAlmostEqual(lr[1], 0.1)
            optimizer.zero_grad()
            optimizer.step()
            with warmup_scheduler.dampening():
                pass

        _test_state_dict(self, warmup_scheduler,
                         lambda: warmup.LinearWarmup(optimizer, warmup_period=10))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytorch-warmup",
    version="0.1.1",
    author="Takenori Yamamoto",
    author_email="yamamoto.takenory@gmail.com",
    description="A PyTorch Extension for Learning Rate Warmup",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Tony-Y/pytorch_warmup",
    packages=['pytorch_warmup'],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.7',
    test_suite='test',
    install_requires=['torch>=1.1']
)



import unittest
import math
import torch
import pytorch_warmup as warmup

from .test_base import _test_state_dict


class TestUntuned(unittest.TestCase):

    def setUp(self):
        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    def test_untuned_linear(self):
        p1 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        p2 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        optimizer = torch.optim.Adam([
                {'params': [p1]},
                {'params': [p2], 'lr': 0.1}
            ], lr=0.5, betas=(0.9, 0.7))
        lr_scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda step: 1.0)
        warmup_scheduler = warmup.UntunedLinearWarmup(optimizer)
        for step in range(1, 11):
            lr = [x['lr'] for x in optimizer.param_groups]
            print(f'{step} {lr}')
            if step < 6:
                self.assertAlmostEqual(lr[0], 0.5 * step / 6)
                self.assertAlmostEqual(lr[1], 0.1 * step / 6)
            else:
                self.assertAlmostEqual(lr[0], 0.5)
                self.assertAlmostEqual(lr[1], 0.1)
            optimizer.zero_grad()
            optimizer.step()
            with warmup_scheduler.dampening():
                lr_scheduler.step()

        _test_state_dict(self, warmup_scheduler,
                         lambda: warmup.UntunedLinearWarmup(optimizer))

    def test_untuned_exponetial(self):
        p1 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        p2 = torch.nn.Parameter(torch.arange(10, dtype=torch.float32).to(self.device))
        optimizer = torch.optim.Adam([
                {'params': [p1]},
                {'params': [p2], 'lr': 0.1}
            ], lr=0.5, betas=(0.9, 0.7))
        lr_scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda step: 1.0)
        warmup_scheduler = warmup.UntunedExponentialWarmup(optimizer)
        for step in range(1, 11):
            lr = [x['lr'] for x in optimizer.param_groups]
            print(f'{step} {lr}')
            self.assertAlmostEqual(lr[0], 0.5 * (1 - math.exp(-step / 3)))
            self.assertAlmostEqual(lr[1], 0.1 * (1 - math.exp(-step / 3)))
            optimizer.zero_grad()
            optimizer.step()
            with warmup_scheduler.dampening():
                lr_scheduler.step()

        _test_state_dict(self, warmup_scheduler,
                         lambda: warmup.UntunedExponentialWarmup(optimizer))
